/***********************************************************************
 *
 * Name:        Plugin_procs.c
 *
 * Description: user and OpenDVE events callbacks               
 *
 * Author:	Janusz Bulawa & Tomasz Stachowiak, NPAC
 *
 * Date:	September 1996
 *
 ***********************************************************************/
 
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/param.h>
#include "VideoProtocol.h"

#include <X11/Intrinsic.h>

#include <Xm/Xm.h>
#include <Xm/MwmUtil.h>
#include <Xm/MenuShell.h>
#include <Xm/RowColumn.h>
#include <Xm/Scale.h>
#include <Xm/MainW.h>

/* H263 display section */
#define GLOBAL
#include "config.h"
#include "tmndec.h"
#include "global.h"

void initdecoder _ANSI_ARGS_((void));

/* private prototypes */
static int getval _ANSI_ARGS_((char *argv[]));

int first[MAX_USERS];

/* end of H263 display section */

extern app;
extern toplevel;
extern start_but;
extern stop_but;
extern shell;

extern localName;
DveUserHandle localHandle;
int    currentUser;
extern unsigned char *frame;
extern char* videobuffer;
extern int numbytes;
extern int writeoffset;
extern int expand;

Window window;
Window win[MAX_USERS];
int dpy_depth;
int depth[MAX_USERS];

extern Display *display;

int totalBytesSent=0;
int totalFramesSent=0;

static unsigned char *localPict[3];
static unsigned char *exlocalPict[3];

Widget MainWindow;

DveSession dvePluginSession=NULL;

extern state;
int tmp_state;

/********************************
 * TIME INTERVAL BETWEEN FRAMES
 ********************************/

int interval = 100;

/********************************/

int numUsers, i;
DveConfParticipant *allUsers, user;
char *wlabel;

int code_type = H263;
int busy = 0;

void start_playing();
void start_conference(void);
void finish(void);
void started(DveSession pluginSession, DveEvent* event);
void HidePlugin();
void ShowPlugin();
void QuitPlugin();
void play(void);
void stop(void);

void SendVideoData(char *data, int len);
void SendMessage(VideoTag messageType, DvePort port, VideoPluginArg *dane);
int SendPluginMessageGeneric(DveSession dveSession, u_int32 protocol, DvePort port,
			u_int32 messageType, int dest, void *data, u_int32 size);
extern SetMessage(const char *msg);

XtIntervalId timerId=0;

  char *data;


/*************************************************
 *
 * Data sending procedures
 *
 *************************************************/

/* Video data sending */
void SendVideoData(char *data, int len)
{
  char *buffer;
  int i;

  totalBytesSent+=len;
  totalFramesSent++;
/*  printf("Sending video data %d bytes...", len);
  for (i=0;i<len;i++) printf("%d ", vdata[i]);
  printf("\n");
  */
  if (len>0)
  {
    SendPluginMessageGeneric(dvePluginSession, VIDEO_PLUGIN_PORT, VIDEO_PLUGIN_PORT,
	  		     code_type, -1, data, len);
  }
  else printf("SendVideoData(): Illegal data length value!\n");

  return;
}

/* Special data sending */
void SendSpecificMessage(VideoTag messageType)
{
   static VideoPluginArg dane;
   char *title;

   switch (messageType) {
   case videoTagPlay:
     dane.frate=10;
     dane.format=SF_QCIF;
     break;
   case videoTagConfirm:
     break;
   case videoTagStop:
     break;
   case videoTagExit:
     break;
   default:
     printf("Unknown request in SendSpecific...!\n");
     break;
   }

   SendMessage(messageType, VIDEO_PLUGIN_PORT, &dane); 
}



/******************************************
 *
 * Received video data processing
 *
 ******************************************/
static void
ProcessVideoData(DveDataPacket* packet, int type)
{
  int i, size, f;

  size = (packet->size)-sizeof(DveDataPacket)+1;
  memset(invideobuffer, 0, MAX_FRAME_SIZE);
  memcpy(invideobuffer, packet->data, size);
  avail=size;
  invideooffset=0;
  initbits();
  if(type == 0)
    {
       if (getheader())
	 {
	    getpicture(&framenum, 0);
	    framenum++;
	 }
    }
  else
    {
       while(busy) {};
       getpicture(&framenum, 1);
    }
}


/********************************************************
 *
 * Main coding and sending procedure, invoked with timer 
 *
 ********************************************************/
void nextFrame()
{

  if(state == PLAY)
    {
      /* frame capturing and compression */
      switch(code_type)
	{
	case H263:
	  numbytes=codeFrame(&frame);
	  break;

	case H261:
	  frame = getFrame(176, 144);
	  numbytes = p64EncodeFrame(frame, videobuffer);
	  break;
	}

      /* mirror image handling */
#ifdef LOCAL
      localPict[0]=frame; 
      localPict[1]=frame+25344;
      localPict[2]=frame+25344+6336;
      currentUser=(int)localHandle-1;
      window=win[currentUser];
      dpy_depth=depth[currentUser];
      if (expand && outtype == T_X11) 
	{
	  interpolate_image(localPict[0], exlocalPict[0],
			    coded_picture_width, coded_picture_height);
	  interpolate_image(localPict[1], exlocalPict[1], 
			    chrom_width, chrom_height);
	  interpolate_image(localPict[2], exlocalPict[2], 
			    chrom_width, chrom_height);
	  dither(exlocalPict);
	}
      else
	{
	  dither(localPict);
	}
#endif
      
      free(frame);
      
      /* compressed frame sending */
      SendVideoData(videobuffer, numbytes);
    } 
  /* setting timer for the same procedure */
  timerId=XtAppAddTimeOut(app,interval,nextFrame,NULL);
}




/**********************************
 *
 * Main panel callbacks
 *
 **********************************/

/*
 * Buttons pushing callbacks
 */
void start_pushed(Widget w, XtPointer data, XmPushButtonCallbackStruct *cbs)
{
  XtSetSensitive(start_but, False);	 
  XtSetSensitive(stop_but, True);  

  state=PLAY;
  SendSpecificMessage(videoTagPlay);

  
  source_format=SF_QCIF;

  if (first[localHandle-1]) 
    {
       initdpy(localHandle - 1);
       first[localHandle-1] = 0; 
    }
   
    
  start_playing();
  SetMessage("Video is running...");
  
  return;
}

void stop_pushed(Widget w, XtPointer data, XmPushButtonCallbackStruct *cbs)
{
  XtSetSensitive(start_but, True);
  XtSetSensitive(stop_but, False);

  state=STOP;

  SetMessage("Video stopped...");
}

/*
 * File menu callback
 */
void file_menu(Widget menu_item, int item_no,
		  XmAnyCallbackStruct *cbs)
{
   switch(item_no)
     {
	case 0:
	SendSpecificMessage(videoTagExit);
	QuitPlugin();
	break;
     }
}


/*
 * Tool menu callback 
 */
void tool_menu(Widget menu_item, int item_no,
		  XmAnyCallbackStruct *cbs)
{
   switch(item_no)
     {
	case 0:
	  if (timerId!=0) 
	    {
	      XtRemoveTimeOut(timerId);
	      timerId=0;
	    }
	  tmp_state = state;
	  state = PANEL;
	  if(shell==NULL) PanelDialog();
	  else XtPopup(shell, XtGrabNone);
	  break;
     }
}


/****************************************************
 *
 * Video control panel callbacks
 *
 ***************************************************/

/* Compression type menu callback */
void
comp_option_cb(w, item_no, cbs)
Widget w;
int item_no;
XmAnyCallbackStruct *cbs;
{
   switch(item_no)
     {
	case 0:
	code_type = H263;
	break;

	case 1:
	code_type = H261;
	break;
     }
}

/* Video size menu callback */
void
size_option_cb(w, item_no, cbs)
Widget w;
int item_no;
XmAnyCallbackStruct *cbs;
{
   switch(item_no)
     {
	case 0:
	printf("QCIF size\n");;
	break;

	case 1:
	printf("CIF size\n");;
	break;
     }
}

/* expand toggle button callback */
void expand_cb(Widget w, char *data, XmToggleButtonCallbackStruct *cbs)
{
  int err, i;


  if(cbs->set)
    {
      expand = 1;
      for(i=0; i<MAX_USERS; i++)
	{
	  if(win[i] != NULL)
	    {
	      err = XDestroyWindow(display, win[i]);
	      initdpy(i);
	    }
	}
    }
	       
   else
     {
       expand = 0;
       for(i=0; i<MAX_USERS; i++)
	 {
	   if(win[i] != 0)
	     {
	       XDestroyWindow(display, win[i]);
	       initdpy(i);
	     }
	 }
     }
}

/* intra mode toggle button callback */
void intra_cb(Widget w, char *data, XmToggleButtonCallbackStruct *cbs)
{
   if(cbs->set)
     setIntraMode();
   else
     resetIntraMode();

}

/*
 * Video panel dialog callbacks
 */
void v_file_menu(Widget menu_item, int item_no,
		  XmAnyCallbackStruct *cbs)
{
   switch(item_no)
     {
	case 0:
	  state = tmp_state;
	  XtPopdown(shell);
	  if(state == PLAY)
	    start_playing();
	  break;
     }
}




void start_playing()
{
  state=PLAY;

  initcapture(176, 144);

  if(code_type == H263)
    {
      frame = getFrame(176, 144);
      codeIntraFrame(frame);
      writeoffset=0;
      SendVideoData(videobuffer, numbytes);
    }

  timerId=XtAppAddTimeOut(app, interval/2, nextFrame, NULL);
}

void started(DveSession dveSession, DveEvent* event)
{
/*  int ilosc, i;
  ConfdStateTag stan;

  ilosc=event->arg.starting.numMembers;
  stan=event->arg.starting.state;
  fprintf(stdout, "I am invited! %d %d\n", ilosc, (int)stan);
  for (i=0;i<ilosc;i++)
    fprintf(stdout, "%s@%s\n", event->arg.starting.members[i].part.userName, event->arg.starting.members[i].part.hostName);
  fflush(stdout);
  */
}


/****************************************
 *
 * DVE functions
 *
 ***************************************/
disconnectPlugin(DveSession dveSession)
{
 
#ifdef XT
   dveUnregisterWithXt(dveSession);
#endif
 
#ifdef XVIEW
   dveUnregisterWithXView(dveSession);
#endif

   dveCloseSession(dveSession);
 
   exit(0);
}


static void PluginControlMsg(DveSession dveSession, DveUserHandle
sender, DveDataPacket* packet) { switch
((DvePluginManagementMsgTag)packet->tag) { case
dvePluginManagementMsgTagQuit: QuitPlugin(); break; case
dvePluginManagementMsgTagHide: HidePlugin(); break; case
dvePluginManagementMsgTagShow: ShowPlugin(); break; } }

static void
VideoControlMsg(DveSession dveSession, DveUserHandle sender,
                 DveDataPacket* packet)
{
  static VideoPluginArg dane;
  memcpy(&dane, packet->data, sizeof(VideoPluginArg));

  switch ((VideoTag)packet->tag)
   { 
      case H263:
	 if ((sender==0)||(sender>numUsers))
	 {
	   printf("Invalid sender handle: %d\n", sender);
	   sender=1;
	 }
	 currentUser = (int)sender-1;
	 window = win[currentUser];
	 dpy_depth = depth[currentUser];
         ProcessVideoData(packet, 0);
	 break;

      case H261:
	 if ((sender==0)||(sender>numUsers))
	 {
	   printf("Invalid sender handle: %d\n", sender);
	   sender=1;
	 }
	 currentUser = (int)sender-1;
	 window = win[currentUser];
	 dpy_depth = depth[currentUser];
         ProcessVideoData(packet, 1);
	 break;

      case videoTagExit:
         if (timerId!=0) XtRemoveTimeOut(timerId);
	 timerId=0;
	 QuitPlugin();
         break;

      case videoTagConfirm:
	break;

      case videoTagPlay:
	 source_format=dane.format;
         if (first[sender-1]) {
           initdpy(sender - 1);
           first[sender-1] = 0;
         }
         SendSpecificMessage(videoTagConfirm);
	 SetMessage("Receiving video...");
         break;

       case videoTagStop:
         break;

       default:
	 printf("Unknown request in VideoControlMsg!\n");
	 break;
   }
    
}

void
HidePlugin()
{
   XtPopdown(toplevel);
}
 
void	
ShowPlugin()
{
   XtPopup(toplevel, XtGrabNone);
}
 
void
QuitPlugin()
{
  /*  printf("Video Plugin: EXIT requested. Exiting...\n");
  printf("Average video data rate: %5.2f bytes per frame\n",
	 (float)totalBytesSent/(float)totalFramesSent);*/
  fflush(stdout);
  disconnectPlugin(dvePluginSession);
  shutdownEncoder();
  exit(0);
}
 
void
introduceYourself()
{
   /*
    * Start up OpenDVE.
    */
   if (!(dvePluginSession=dveOpenSession(VIDEO_PLUGIN_PORT)))
   {
      fprintf(stderr, "Initialization of OpenDVE Failed...exiting.\n\7");
      switch (dveError)
	{
	case dveErrOutOfMemory:
	  fprintf(stderr, "Error Condition: [Low memory.]\n");
	  break;
	  
	case dveErrNotFound:
	  fprintf(stderr, 
		  "Error Condition: [Conference Engine not found.]\n");
	  break;
	  
	case dveErrConnRefused:
	  fprintf(stderr, 
		  "Error Condition: [Engine refused connection.]\n");
	  break;
	default:
	  fprintf(stderr, "Unknown Error Condition.\n");
	  break;
	}
       exit(0);
   }

   dveSetUserCallback(dvePluginSession, VIDEO_PLUGIN_PORT, VideoControlMsg);
   dveSetUserCallback(dvePluginSession, dvePluginManagementProtocol,
		      PluginControlMsg);

   dveSetEventCallback(dvePluginSession, dveEventTagStarting,started);   
    /*
    * Register this event processing with the Xt toolkit.  Without this
    * call OpenDVE events will not be received.
    */
   
   if (dveRegisterWithXt(dvePluginSession, app)<0)
   {
      fprintf(stderr, "Unable to register with Xt loop.");
      dveCloseSession(dvePluginSession);
      exit(0);
   }

   dveReportPresence(dvePluginSession);

   localName=(char*)dveMalloc(64);
   localName=dveGetLocalHostName(dvePluginSession);
   localHandle=dveGetLocalUserHandle(dvePluginSession);
/*   printf("Local User's handle: %d\n", localHandle);*/
   fflush(stdout);
   
   allUsers=(DveConfParticipant *)dveMalloc(MAX_USERS *
		  		            sizeof(DveConfParticipant));
   numUsers=dveGetAllMembers(dvePluginSession, allUsers, MAX_USERS);

   if (numUsers>MAX_USERS)
   {
     printf("Too many users for videokonference! Exiting...\n");
     fflush(stdout);
     exit(-1);
   }
     
/*   for (i=0;i<numUsers;i++)
   {
     printf("User %d handle %d\n", i, allUsers->userHandle);
     allUsers++;
   }
   printf("There is %d participants in teleconference\n", numUsers);*/
   dveFree(allUsers);
   
  for(i=0; i<MAX_USERS; i++)
    first[i]=1;

  quiet=1; 
  ld = &base; 

  framenum = 0;
  temp_ref = 0;
  prev_temp_ref= -1;
  state=STOP; 
  outputname = "";
  videobuffer=(char*)dveMalloc(MAX_FRAME_SIZE);
  invideobuffer=(char*)dveMalloc(MAX_FRAME_SIZE);
  vdata = dveMalloc(MAX_FRAME_SIZE);
  avail=0;

   return;
}

/*
 * Notify other conference members of our current settings.  We have
 * either made a change or have received a request to provide status.
 */
void
SendMessage(VideoTag messageType, DvePort port, VideoPluginArg *dane)
{

   SendPluginMessageGeneric(dvePluginSession, port, port, messageType, -1, dane, sizeof(*dane));

   return;
}

/*
 * Passses data from this plugin to other conferenced members.
 */
int
SendPluginMessageGeneric(DveSession dveSession, u_int32 protocol, DvePort port,
			u_int32 messageType, int dest, void *data, u_int32 size)
{
   DveDataPacket *packet;
   DveConfParticipant member;
   int returnCode,i;

/*   printf("Size of data to send: %d\n", size);*/
   
   packet = (DveDataPacket*) dveMallocDataPacket(dveSession, size);
   packet->tag = (u_int32)messageType;
   packet->pluginProtocol = protocol;
   packet->destPort = port;

/*   printf("SendGeneric to port %d\n", packet->destPort);*/
   DvePartListClrAll(&(packet->destParts));

   if (dest < 0)
   {
      DvePartListSetAll(&(packet->destParts));
   }
   else
   {
      DvePartListSet(dest,&(packet->destParts));
   }

   if (size > 0)
      memcpy(packet->data, data, size);

/*   if (messageType==videoTagData)
    {
      printf("\nDane w pakiecie:\n");
      for(i=0;i<size;i++) printf("%d ", packet->data[i]);
      printf("\n");
    }
    */
   returnCode = dveSendDataPacket(dveSession, packet);

   dveFreeDataPacket(packet); 
   return(returnCode);
}




/* H263 display section */
initdpy(int userno)
{
  MainWindow = XtWindow(toplevel);

  dveGetOneMember(dvePluginSession, (DveUserHandle)(userno+1), &user);
  wlabel=(char*)dveMalloc(128);
  wlabel=user.hostName;
  init_display("", wlabel, (int)MainWindow, userno);
  dveFree(wlabel);
}


initexpand()
{
  int cc, k, size;

  for (cc=0; cc<3; cc++) 
    {
      if (cc==0)
	size = coded_picture_width*coded_picture_height*4;
      else
	size = chrom_width*chrom_height*4;
    
      if (!(exlocalPict[cc] = (unsigned char *)malloc(size)))
	error("malloc failed\n");

      for (k=0;k<MAX_USERS;k++) 
	{
	  if (!(exnewframe[k][cc] = (unsigned char *)malloc(size)))
	    error("malloc failed\n");
	}
    }
}

void initdecoder(void)
{
  int i, k, cc, size;
  FILE *cleared;

  source_format = SF_QCIF;

  expand = 0;
  outtype = T_X11;

  /* clip table */
  if (!(clp=(unsigned char *)malloc(1024)))
    error("malloc failed\n");

  clp += 384;

  for (i=-384; i<640; i++)
    clp[i] = (i<0) ? 0 : ((i>255) ? 255 : i);

  /* MPEG-1 = TMN parameters */
  matrix_coefficients = 5;

  switch (source_format) {
    case (SF_SQCIF):
      horizontal_size = 128;
      vertical_size = 96;
      break;
    case (SF_QCIF):
      horizontal_size = 176;
      vertical_size = 144;
      break;
    case (SF_CIF):
      horizontal_size = 352;
      vertical_size = 288;
      break;
    case (SF_4CIF):
      horizontal_size = 704;
      vertical_size = 576;
      break;
    case (SF_16CIF):
      horizontal_size = 1408;
      vertical_size = 1152;
      break;
    default:
      printf("ERROR: Illegal input format\n");
      exit(-1);
      break;
  }


  mb_width = horizontal_size/16;
  mb_height = vertical_size/16;
  coded_picture_width = horizontal_size;
  coded_picture_height = vertical_size;
  chrom_width =  coded_picture_width>>1;
  chrom_height = coded_picture_height>>1;
  blk_cnt = 6;

  for (cc=0; cc<3; cc++) {
    if (cc==0)
      size = coded_picture_width*coded_picture_height;
    else
      size = chrom_width*chrom_height;

    for (k=0;k<MAX_USERS;k++) 
      {
	if (!(refframe[k][cc] = (unsigned char *)malloc(size)))
	  error("malloc failed\n");
	if (!(oldrefframe[k][cc] = (unsigned char *)malloc(size)))
	  error("malloc failed\n");
	if (!(bframe[k][cc] = (unsigned char *)malloc(size)))
	  error("malloc failed\n");
      }
  }

  for (cc=0; cc<3; cc++) {
    if (cc==0) {
      size = (coded_picture_width+64)*(coded_picture_height+64);
      for (k=0;k<MAX_USERS;k++)
	{
	  if (!(edgeframeorig[k][cc] = (unsigned char *)malloc(size)))
	    error("malloc failed\n");
	  edgeframe[k][cc] = edgeframeorig[k][cc] + (coded_picture_width+64) * 32 + 32;
	}
    }
    else {
      size = (chrom_width+32)*(chrom_height+32);
      for (k=0;k<MAX_USERS;k++) 
	{
	  if (!(edgeframeorig[k][cc] = (unsigned char *)malloc(size)))
	    error("malloc failed\n");
	  edgeframe[k][cc] = edgeframeorig[k][cc] + (chrom_width+32) * 16 + 16;
	}
    }
  }

  /*  if (expand) 
    {*/
      initexpand();
      /*    }*/

  /* IDCT */
  if (refidct)
    init_idctref();
  else
    init_idct();

  for(i=0; i<MAX_USERS; i++)
    {
      win[i] = NULL;
    }

  init_global_dpy();
}

void error(text)
char *text;
{
  fprintf(stderr,text);
  exit(1);
}

/* trace output */
void printbits(code,bits,len)
int code,bits,len; 
{
  int i;
  for (i=0; i<len; i++)
    printf("%d",(code>>(bits-1-i))&1);
}


static int getval(argv)
char *argv[];
{
  int val;

  if (sscanf(argv[1]+2,"%d",&val)!=1)
    return 0;

  while (isdigit(argv[1][2]))
    argv[1]++;

  return val;
}


/* end of H263 display section */






